from typing import List, Tuple, Union
from pydantic import BaseModel
from sklearn.model_selection import train_test_split


class YAFALDataset(BaseModel):
    """Dataset model to handle the required transformations in YAFAL!
    """
    samples: List[str] = []
    labels: List[List[int]] = []
    is_corrupted: List[int] = []

    def add_sample(self, text: str, sample_labels: List[Union[int, str]], is_corrupted: int = 0) -> None:
        """ Adds a sample to the dataset

        Parameters
        ----------
        text : str
            Sample to add

        sample_labels : List[Union[int, str]]
            List of the labels associated to this sample. The list should contain only one item for single-label
            settings, else, more.

        is_corrupted: int, default = 0
            If the sample is corrupted (wrong label) or not. 0 if OK, 1 if corrupted.

        """
        self.samples.append(text)
        self.labels.append(sample_labels)
        self.is_corrupted.append(is_corrupted)

    @property
    def label_list(self):
        label_list = []
        for labels in self.labels:
            label_list += labels
        return sorted(list(set(label_list)))

    def __iter__(self) -> Tuple[str, List[int], int]:
        for sample, label, is_corrupted in zip(self.samples, self.labels, self.is_corrupted):
            yield sample, label, is_corrupted

    def __len__(self):
        return len(self.samples)

    def split_indexes(self, ratio: float) -> Tuple[List[int], List[int]]:
        """ Split the dataset according to a given ratio and returns the split indexes

        """
        train_indexes, validation_indexes, labels_train, label_validations = \
            train_test_split(range(len(self.samples)), self.labels, test_size=ratio)
        return train_indexes, validation_indexes

    def __getitem__(self, index: int):
        return self.samples[index], self.labels[index], self.is_corrupted[index]




import os

from yafal import YAFALDataset, YAFALRegression
from yafal.corruption import DatasetCorruptor
import pandas as pd
from sklearn.metrics import roc_auc_score, accuracy_score, precision_score, recall_score

current_dir = os.path.dirname(os.path.abspath(__file__))
train_csv_path = os.path.join(current_dir, 'data', 'train.csv')
test_csv_path = os.path.join(current_dir, 'data', 'test.csv')

train_data = pd.read_csv(train_csv_path)
test_data = pd.read_csv(test_csv_path)

topic_mapping = {1: 'world',
                 2: 'sports',
                 3: 'business',
                 4: 'science'}

train_yafal_dataset = YAFALDataset()
total_samples = 10000  # To make everything faster
for i, sample in train_data.iterrows():
    index_label = sample['Class Index']
    string_label = topic_mapping[index_label]
    news_article = f"{sample['Title']}. {sample['Description']}"
    train_yafal_dataset.add_sample(
        text=news_article,
        sample_labels=[string_label]
    )
    # For speedup
    if i == total_samples:
        break

# Train our model
print("Let's train our YAFAL Regressor")
yafal_regression = YAFALRegression(label_encoder_method="semantic", verbose=True, max_epochs=10)
yafal_regression.fit(train_yafal_dataset,
                     descriptors={"world": ["politics", "international", "global", "world"],
                                  "sports": ["soccer", "basketball", "sport", "workout", "gym"],
                                  "business": ["economics", "trade", "money", "business"],
                                  "science": ["science", "technology", "innovation", "research"]})

# Save the model

yafal_regression.save("YAFAL_ag_semantic_model.pck")

# Test the model - Load a new YAFAL dataset for the test
test_yafal_dataset = YAFALDataset()
total_test_samples = 2000
for i, sample in train_data.iterrows():
    index_label = sample['Class Index']
    string_label = topic_mapping[index_label]
    news_article = f"{sample['Title']}. {sample['Description']}"
    test_yafal_dataset.add_sample(
        text=news_article,
        sample_labels=[string_label]
    )
    if i == total_test_samples:
        break

# Corrupt the dataset
corruptor = DatasetCorruptor()
corrupted_test_dataset = corruptor.transform(test_yafal_dataset)

# Predict with our model
y_true, y_predicted = [], []
for sample, labels, is_corrupted in corrupted_test_dataset:
    result = yafal_regression.predict([sample], labels=labels)
    y_true.append(is_corrupted)
    y_predicted.append(result.item())

y_predicted_class = [0 if predict < 0.5 else 1 for predict in y_predicted]

print('Accuracy score: {}'.format(accuracy_score(y_true, y_predicted_class)))
print('Precision score: {}'.format(precision_score(y_true, y_predicted_class)))
print('Recall score: {}'.format(recall_score(y_true, y_predicted_class)))
print('ROC AUC score: {}'.format(roc_auc_score(y_true, y_predicted)))



import random

import numpy as np

from yafal.label import YAFALLabelEncoder


def test_label_encoder_binary_single_label():
    encoder = YAFALLabelEncoder()
    # Test 1: use integers as labels
    label_list = list(range(10))*2
    encoder.fit(label_list)
    assert set(encoder.label_indexes) == set(label_list)

    # Test1.1: Transform each label individually
    for i, label in enumerate(range(10)):
        result = encoder.transform([label])
        result_to_compare = np.zeros((1, len(label_list)))
        result_to_compare[0, i] = 1
        assert result.all() == result_to_compare.all()

    # Test 1.2: Transform several items at once
    batch_result = encoder.transform(list(range(10)))
    assert batch_result.all() == np.eye(10).all()

    # Test 1.3: test the empty result
    assert np.zeros((0, 10)).all() == encoder.transform([]).all()
    # Test 1.4: test some unseen label - this should crash
    for unseen in [-1, len(label_list)+2]:
        try:
            encoder.transform([unseen])
        except KeyError:
            # This should happen
            pass
        except Exception:
            raise Exception("The label encoder does not fail when unseen labels are given")

    # Test 2: use texts as labels
    encoder = YAFALLabelEncoder()
    labels = [f"label-{i}" for i in range(10)]*2
    encoder.fit(labels)
    # Test that there are 10 total labels
    assert set(encoder.label_indexes) == set(list(range(10)))
    # Test that every label is in the label map
    for label in labels:
        assert label in encoder.label_map

    for i, label in enumerate(labels):
        result = encoder.transform([label])
        result_to_compare = np.zeros((1, len(label_list)))
        result_to_compare[0, i] = 1
        assert result.all() == result_to_compare.all()
    batch_result = encoder.transform(labels)
    assert batch_result.all() == np.eye(10).all()


def test_label_encoder_binary_multi_label():
    encoder = YAFALLabelEncoder()
    # Test 1: use integers as labels
    label_list = list(range(10))
    encoder.fit(label_list)
    assert set(encoder.label_indexes) == set(label_list)

    # Test1.1: Transform each label individually
    samples, transformations = [], []
    for try_number in range(1000):
        items_to_sample = random.choice(list(range(1, 10)))
        selected_labels = random.sample(label_list, k=items_to_sample)
        expected_result = np.zeros((1, 10))
        expected_result[0, selected_labels] = 1
        assert encoder.transform([selected_labels]).all() == expected_result.all()
        samples.append(selected_labels)
        transformations.append(expected_result)

    assert encoder.transform(samples).all() == np.vstack(transformations).all()

    # Test 2: use texts as labels
    encoder = YAFALLabelEncoder()
    labels = [f"label-{i}" for i in range(10)] * 2
    encoder.fit(labels)
    # Test that there are 10 total labels
    assert set(encoder.label_indexes) == set(list(range(10)))
    # Test that every label is in the label map
    for label in labels:
        assert label in encoder.label_map

    samples, transformations = [], []
    for try_number in range(1000):
        items_to_sample = random.choice(list(range(1, 10)))
        selected_labels = random.sample(labels, k=items_to_sample)
        selected_label_indexes = [labels.index(selected_label) for selected_label in selected_labels]
        expected_result = np.zeros((1, 10))
        expected_result[0, selected_label_indexes] = 1
        assert encoder.transform([selected_labels]).all() == expected_result.all()
        samples.append(selected_labels)
        transformations.append(expected_result)

    assert encoder.transform(samples).all() == np.vstack(transformations).all()


def test_label_encoder_semantic_single_label():
    encoder = YAFALLabelEncoder()
    labels = [f"label-{i}" for i in range(10)]
    encoder.fit(labels,
                label_descriptors={label: [f"desc-{j}" for j in range(i)] for i, label in enumerate(labels)})
    assert set(encoder.label_indexes) == set(range(len(labels)))
    # The first label has no descriptors, so the check should not be ok
    assert not encoder.validate_descriptions()
    # Let's update the descriptor of the first label
    encoder.update_descriptors(
        label_identifier="label-0",
        descriptors=["desc-0"]
    )
    # Now everything should be OK
    assert encoder.validate_descriptions()
    for label_index in range(1, len(labels)):
        label = labels[label_index]
        supposed_descriptor = [' [SEP] '.join([f"desc-{j}" for j in range(label_index)])]
        encoder_result = encoder.describe([label], max_descriptors=10)
        assert encoder_result == supposed_descriptor

    # Check that the max_descriptors parameter works
    for i in range(1, len(labels)):
        descriptions = encoder.describe(labels, max_descriptors=i)
        for descr in descriptions:
            assert len(descr.split(' [SEP] ')) <= i




class InvalidMethodException(Exception):
    pass


class DescriptorsRequiredException(Exception):
    pass


class LabelInconsistencyException(Exception):
    pass


class LabelTypeException(Exception):
    pass



from yafal import YAFALDataset
from yafal.corruption import DatasetCorruptor
import random

random.seed(42)


def _get_corrupted_dataset():
    int_label_dataset = YAFALDataset()
    text_label_dataset = YAFALDataset()
    for i in range(5):
        int_label_dataset.add_sample(
            "Label 1 data", sample_labels=[1]
        )
        int_label_dataset.add_sample(
            "Label 0 data", sample_labels=[0]
        )
        # Text dataset
        text_label_dataset.add_sample(
            "Label 1 data", sample_labels=["Yes"]
        )
        text_label_dataset.add_sample(
            "Label 0 data", sample_labels=["No"]
        )
    return int_label_dataset, text_label_dataset


_get_corrupted_dataset()


def test_no_corruption():
    int_label_dataset, text_label_dataset = _get_corrupted_dataset()
    corruptor = DatasetCorruptor(probability=0)
    int_corrupt = corruptor.transform(int_label_dataset)
    text_corrupt = corruptor.transform(text_label_dataset)
    for sample_1, sample_2 in zip(int_label_dataset, int_corrupt):
        assert sample_1 == sample_2

    for sample_1, sample_2 in zip(text_label_dataset, text_corrupt):
        assert sample_1 == sample_2


def test_keep_original_parameter():
    int_label_dataset, text_label_dataset = _get_corrupted_dataset()
    corruptor = DatasetCorruptor(probability=1.1, keep_original=False)
    int_corrupt = corruptor.transform(int_label_dataset)
    text_corrupt = corruptor.transform(text_label_dataset)
    assert len(int_corrupt.samples) == len(int_label_dataset.samples)
    assert len(text_corrupt.samples) == len(text_label_dataset.samples)

    for sample_1, sample_2 in zip(int_label_dataset, int_corrupt):
        # The labels should be different
        assert sample_1[1] != sample_2[1] and sample_1[2] != sample_2[2]
        # Check that the label types are kept (no string converted to integer or vice-versa)
        assert isinstance(sample_1[1][0], type(sample_2[1][0]))

    for sample_1, sample_2 in zip(text_label_dataset, text_corrupt):
        # The labels should be different
        assert sample_1[1] != sample_2[1] and sample_1[2] != sample_2[2]
        # Check that the label types are kept (no string converted to integer or vice-versa)
        assert isinstance(sample_1[1][0], type(sample_2[1][0]))

    corruptor = DatasetCorruptor(probability=1.1, keep_original=True)
    int_corrupt = corruptor.transform(int_label_dataset)
    text_corrupt = corruptor.transform(text_label_dataset)
    assert len(int_corrupt.samples) == 2*len(int_label_dataset.samples)
    assert len(text_corrupt.samples) == 2*len(text_label_dataset.samples)

    # Check that the original items exists on the corrupted datasets

    assert {str(sample) for sample in int_label_dataset}.issubset({str(corrupted) for corrupted in int_corrupt})
    assert {str(sample) for sample in text_label_dataset}.issubset({str(corrupted) for corrupted in text_corrupt})



"""
Implements the Semantic Label Encoder
"""
import typing
from typing import List, Union

import numpy as np


class YAFALLabelEncoder:
    """Performs the semantic encoding of the labels for the YAFAL regressors
    """

    def __init__(self, ):
        """
        Initializes the YAFAL Label Encoder Class
        """

        self.label_indexes = []
        self.label_map = {}

    @staticmethod
    def __check_label_type(label_list: List[Union[int, str]]) -> None:
        label_types = set([type(label) for label in label_list])
        if len(label_types) != 1:
            raise Exception("Only one data type is expected for labels, received: {}".format(set(label_types)))
        label_type = label_types.pop()
        if label_type not in [int, str]:
            raise Exception("Label type needs to be int or str, received: {}".format(label_type))

    def fit(self, labels: List[Union[int, str]],
            label_descriptors=None) -> None:
        """
        Builds the YAFAL Label Encoder with the required labels and optionally with the label descriptors

        Parameters
        ----------

        labels: list, list of integers of strings
            List of labels to train the YAFAL Label Encoder

        label_descriptors: optional, dictionary with the label descriptors
            Dictionary of type ``{label: [descriptor_text, descriptor_text2, ...]}``

        """
        self.__check_label_type(labels)
        for label in labels:
            if not self.label_map.get(label):
                # The label does not exists
                label_index = len(self.label_indexes)
                self.label_indexes.append(label_index)
                self.label_map[label] = {
                    "index": label_index,
                    "descriptors": []
                }
        if label_descriptors:
            for label_identifier in label_descriptors:
                if label_identifier in self.label_map:
                    self.label_map[label_identifier]["descriptors"] = label_descriptors[label_identifier]
                else:
                    raise Exception('Label "{}" not found when updating descriptors.'.format(label_identifier))

    def transform(self, labels: List[Union[int, str, List[Union[str, int]]]]) -> np.ndarray:
        """
        Given a list of labels, composed either of integers, strings or lists (for multilabel)
        and returns the numpy array representation.

        Parameters
        ----------

        labels: list
            List of labels, they can be of type int, str or nested for multi-label classifiers
            ``labels = [1, 0 , 0, 1, 1]``

        Returns
        -------

        label_array : ndarray of shape (n_samples, n_labels)
            Label array
        """
        label_transform = np.zeros((len(labels), len(self.label_indexes)))
        for i in range(len(labels)):
            if isinstance(labels[i], list):
                for sub_label in labels[i]:
                    label_index = self.label_map[sub_label]["index"]
                    label_transform[i, label_index] = 1
            else:
                label_index = self.label_map[labels[i]]["index"]
                label_transform[i, label_index] = 1
        return label_transform

    def describe(self, labels: List[Union[int, str]], separator_token: str = " [SEP] ",
                 max_descriptors: int = 5) -> typing.List[str]:
        """Creates the description of the given labels for semantic encoding in LLM.

        .. warning::
            Currently it does not support multi-label description. I still have my doubts if in a multi-label
            setup if a micro or macro corruption method should be taken into account.

        Parameters
        ----------

        labels : list
            List of labels to encode:

                ``labels = [1, 0 , 0, 1, 1]``

                ``labels = ["hi", "bye" , "hi", "bye", "hi"]``

        separator_token : str, default=" [SEP] "
            Separator token to join the different descriptions given for each label. Currently, the separator used for
            BERT is used.

        max_descriptors : int, default=5
            Amount of descriptors to use to describe a label

        Returns
        -------
        list
            List of strings describing each one of the labels

        """
        if max_descriptors < 1:
            raise Exception("At least one descriptor needs to be employed (max_descriptions=1)")
        texts_to_return = []
        for label in labels:
            texts_to_return.append(
                separator_token.join(self.label_map[label]["descriptors"][:max_descriptors])
            )
        return texts_to_return

    def update_descriptors(self, label_identifier: Union[int, str], descriptors: List[str]) -> None:
        """
        Updates the descriptors of a label identifier

        Parameters
        ----------

        :param label_identifier: str
            Identifier of the label to update, it must be available in the label map

        :param descriptors: list, List of strings used to describe the label


        """
        if label_identifier not in self.label_map:
            raise Exception('Label id "{}" not found in the label map.'.format(label_identifier))
        self.label_map[label_identifier]["descriptors"] = descriptors

    def validate_descriptions(self) -> bool:
        """
        Checks that all the labels have the required semantic description

        Returns
        -------

        :return: bool,
            If the descriptions are valid or not

        """
        if not self.label_map:
            return False
        for label in self.label_map:
            if not self.label_map[label]["descriptors"]:
                return False
        return True



from typing import Tuple

from transformers import AutoModel, AutoTokenizer
import torch
from torch import nn, Tensor


class BinaryYAFAL(nn.Module):
    def __init__(self, label_encoding_size: int,
                 pretrained_model_name: str = "distilbert-base-uncased",
                 mlp_layer_sizes: Tuple[int] = (150, ),
                 llm_output_shape: int = 768
                 ):
        super(BinaryYAFAL, self).__init__()
        self.__model_name = pretrained_model_name
        self.tokenizer = AutoTokenizer.from_pretrained(self.__model_name)
        self.bert = AutoModel.from_pretrained(self.__model_name)
        self.linear1 = nn.Linear(llm_output_shape+label_encoding_size, mlp_layer_sizes[0])
        self.__llm_size = llm_output_shape
        self.linear_layers = [
        ]
        for i in range(1, len(mlp_layer_sizes)):
            self.linear_layers.append(
                nn.Linear(mlp_layer_sizes[i-1], mlp_layer_sizes[i])
            )

        self.final_layer = nn.Linear(mlp_layer_sizes[-1], 1)

    def _get_classifier_token_embedding(self, sequence_output):
        return sequence_output[:, 0, :].view(-1, self.__llm_size)

    def forward(self, text, label_encoding: Tensor):
        tokenization_output = self.tokenizer(text, return_tensors='pt', padding=True, truncation=True, max_length=50,
                                             add_special_tokens=True)

        sequence_output = self.bert(**tokenization_output)

        classify_tensor = self._get_classifier_token_embedding(sequence_output.last_hidden_state)

        # Sequence_output has the following shape: (batch_size, sequence_length, 768)
        # Join the output with the label encoding representation
        concatenate_layer = torch.cat((classify_tensor, label_encoding), 1)
        linear_output = self.linear1(concatenate_layer)
        for linear_layer in self.linear_layers:
            tanh_result = nn.Tanh()
            tanh_output = tanh_result(linear_output)
            linear_output = linear_layer(tanh_output)
        tanh_result = nn.Tanh()
        tanh_output = tanh_result(linear_output)
        linear2_output = self.final_layer(tanh_output)
        sigmoid_result = nn.Sigmoid()
        return sigmoid_result(linear2_output)



import random
import typing

from yafal.models import YAFALDataset


class DatasetCorruptor:
    def __init__(self, probability: float = 0.3, keep_original: bool = True,
                 n_rounds: int = 1, max_labels_to_corrupt: int = -1):
        """
        Class used to corrupt / flip the labels of an original YAFAL dataset.

        Parameters
        ----------

        probability: float, default=0.3
            Probability to select a sample to flip the labels from it.

        keep_original: bool, default=True
            If we want to keep the original sample after the corruption or not.

        n_rounds: int, default=1
            How many times do we want to iterate over this dataset to corrupt it.

        max_labels_to_corrupt: int, default=-1
            Mostly for multi-label tasks, how many labels do we want to corrupt if the sample is selected.
            For example, if the sample has 4 labels, and if 'max_labels_to_corrupt'=3, three will be flipped.
            If -1 is selected, all the labels will be flipped/corrupted randomly.

        """
        self._p = probability
        self._keep_original = keep_original
        self._n_rounds = n_rounds
        self._max_labels_to_corrupt = max_labels_to_corrupt

    def flip_labels(self, original_labels: typing.List[int], available_labels: typing.List[int]) -> typing.List[int]:
        """
        Flip the original labels using the set of available labels to that ent

        Parameters
        ----------

        original_labels: List[int]
            Original labels to flip

        available_labels: List[int]
            Labels that can be used to flip the original labels

        Returns
        -------
        List[int]
            List of new (flipped) labels
        """
        if self._max_labels_to_corrupt < 0:
            # Corrupt all the labels
            num_labels_to_corrupt = len(original_labels)
        else:
            num_labels_to_corrupt = min(len(original_labels), self._max_labels_to_corrupt)

        selected_labels = random.sample(original_labels, k=num_labels_to_corrupt)

        # Initialise the new labels with the ones that are not selected to be corrupted
        new_labels = [label for label in original_labels if label not in selected_labels]

        for label_to_corrupt in selected_labels:
            corrupted_label = random.choice([possible_label for possible_label in available_labels if
                                             possible_label != label_to_corrupt])
            new_labels.append(corrupted_label)

        return list(set(new_labels))

    def transform(self, dataset: YAFALDataset) -> YAFALDataset:
        """
        Transforms the input dataset into a corrupted one

        Parameters
        ----------

        dataset: YAFALDataset
            Original, non-corrupted, dataset

        Returns
        -------

        YAFALDataset:
            Corrupted dataset using the above one
        """
        corrupted_dataset = YAFALDataset()

        for epoch in range(self._n_rounds):

            for sample, label_list, _ in dataset:
                if self.__flip_sample():
                    new_labels = self.flip_labels(label_list, available_labels=dataset.label_list)
                    corrupted_dataset.add_sample(
                        text=sample,
                        sample_labels=new_labels,
                        is_corrupted=1
                    )
                    if self._keep_original:
                        corrupted_dataset.add_sample(
                            text=sample,
                            sample_labels=label_list,
                            is_corrupted=0
                        )
                else:
                    corrupted_dataset.add_sample(
                        text=sample,
                        sample_labels=label_list,
                        is_corrupted=0
                    )

        return corrupted_dataset

    def __flip_sample(self) -> bool:
        return random.random() < self._p



import copy
import os

from yafal import YAFALDataset, YAFALRegression
from yafal.corruption import DatasetCorruptor
from yafal.exceptions import InvalidMethodException, DescriptorsRequiredException

train_yafal_dataset = YAFALDataset()
for i in range(50):
    train_yafal_dataset.add_sample(
        text="Marvin is a happy robot",
        sample_labels=[0]
    )
    train_yafal_dataset.add_sample(
        text="The ultimate answer is 42",
        sample_labels=[1]
    )

test_yafal_dataset = copy.deepcopy(train_yafal_dataset)

semantic_yafal_dataset = YAFALDataset()
for i in range(50):
    semantic_yafal_dataset.add_sample(
        text="Marvin is a happy robot",
        sample_labels=["robot"]
    )
    semantic_yafal_dataset.add_sample(
        text="The ultimate answer is 42",
        sample_labels=["life"]
    )

semantic_test_dataset = copy.deepcopy(semantic_yafal_dataset)


def test_train_binary_model():
    # Train our model
    yafal_regression = YAFALRegression(label_encoder_method="binary", verbose=True, max_epochs=5,
                                       mlp_layer_sizes=(150, 75))
    yafal_regression.fit(train_yafal_dataset)
    # Corrupt the dataset
    corruptor = DatasetCorruptor()
    corrupted_test_dataset = corruptor.transform(test_yafal_dataset)

    # Predict with our model
    for sample, labels, is_corrupted in corrupted_test_dataset:
        _ = yafal_regression.predict([sample], labels=labels)

    yafal_regression.save("test.pck")
    os.remove("test.pck")


def test_train_semantic_model():
    # Train our model
    yafal_regression = YAFALRegression(label_encoder_method="semantic", verbose=True, max_epochs=5,
                                       mlp_layer_sizes=(150, 75))
    yafal_regression.fit(semantic_yafal_dataset, descriptors={"robot": ["Marvin"],
                                                              "life": ["meaning of life"]})

    # Corrupt the dataset
    corruptor = DatasetCorruptor()
    corrupted_test_dataset = corruptor.transform(semantic_test_dataset)

    # Predict with our model
    for sample, labels, is_corrupted in corrupted_test_dataset:
        _ = yafal_regression.predict([sample], labels=labels)


def test_trigger_no_description_exceptions():
    try:
        YAFALRegression(label_encoder_method="potatoes", verbose=True, max_epochs=5)
        raise Exception
    except InvalidMethodException:
        pass

    try:
        yafal_regression = YAFALRegression(label_encoder_method="semantic", verbose=True, max_epochs=5)
        yafal_regression.fit(semantic_yafal_dataset, descriptors={"robot": ["Marvin"]})
        raise Exception
    except DescriptorsRequiredException:
        pass

    try:
        yafal_regression = YAFALRegression(label_encoder_method="semantic", verbose=True, max_epochs=5)
        yafal_regression.fit(semantic_yafal_dataset)
        raise Exception
    except DescriptorsRequiredException:
        pass



import pickle
import pprint
from typing import Callable, Dict, Union, List, Optional, Tuple

import numpy as np
from torch.utils.data import Dataset, DataLoader, RandomSampler
import torch
from torch import nn

from yafal.corruption import DatasetCorruptor
from yafal.label import YAFALLabelEncoder
from yafal.models import YAFALDataset
from .binary import BinaryYAFAL
from .semantic import SemanticYAFAL
from ..exceptions import InvalidMethodException, DescriptorsRequiredException


class YAFALTorchDataset(Dataset):
    def __init__(self, dataset: YAFALDataset,
                 label_encoder: Callable,
                 label_transform: str = "binary",
                 ):
        """
        Torch handler for the YAFAL dataset class to be used by a Torch data handler

        Parameters
        ----------

        :param dataset:
        :param label_transform: Callable
            method to transform
        """

        self.dataset = dataset
        self.label_transform = label_transform
        self.label_encoder = label_encoder

    def __len__(self):
        return len(self.dataset.samples)

    def add_sample(self, *args, **kwargs):
        return self.dataset.add_sample(*args, **kwargs)

    def __getitem__(self, index):
        sample = self.dataset.samples[index]
        if self.label_transform == "binary":
            original_label_transform = torch.FloatTensor(
                self.label_encoder(self.dataset.labels[index])[0]
            )
        else:
            original_label_transform = self.label_encoder(self.dataset.labels[index])[0]
        label_to_predict = torch.FloatTensor([self.dataset.is_corrupted[index]])
        return sample, original_label_transform, label_to_predict


class YAFALRegression:
    def __init__(self, label_encoder_method: str,
                 corrupt_percentage: float = 0.3,
                 keep_original: bool = True,
                 n_rounds: int = 1,
                 max_labels_to_corrupt: int = -1,
                 validation_corpus_ratio: float = 0.2,
                 max_epochs: int = 4,
                 verbose: bool = False,
                 device: str = "cpu",
                 batch_size: int = 64,
                 freeze_label_encoder: bool = True,
                 freeze_sequence_encoder: bool = True,
                 early_stopping: bool = True,
                 patience: int = 5,
                 **yafal_kwargs
                 ):
        """

        Parameters
        ----------

        label_encoder_method: str
            String in "binary" or "semantic"

        corrupt_percentage: float, default=0.3
            The amount of noise to add to the original dataset

        keep_original: bool, default=True
            If we want to keep the original sample once it is corrupted

        n_rounds: int, default=1
            Number of rounds to corrupt the dataset

        max_labels_to_corrupt: int, default=-1
            Maximum amount of labels to corrupt from the sample (for multi-label). If -1, all labels will be flipped.

        validation_corpus_ratio: float, default=0.2
            Percentage of the corpus to use for validation

        max_epochs: int, default=4
            Number of maximum epochs to train the network

        verbose: bool, default=False
            Print debugging info during training or not

        device: str, default="cpu"
            The device to use for training

        batch_size: int, default=64
            Batch size for training

        freeze_label_encoder: bool, default=True
            If we want to freeze the LLM that encodes the model for training

        freeze_sequence_encoder: bool, default=True
            If we want to freeze the LLM that encodes the sequence for training

        early_stopping: bool, default=True
            If we want to stop the training early if the metrics are not being improved in the validation set

        patience: int, default=5
            Number of rounds to wait where the validation set scores do not improved before stopping

        **yafal_kwargs: kwargs
            Keywords arguments that are passed to the YAFALRegressors' underlying model

        """

        self._verbose = verbose
        self._yafal_kwargs = yafal_kwargs
        self._label_encoder = YAFALLabelEncoder()
        self._dataset_corruptor = DatasetCorruptor(probability=corrupt_percentage, keep_original=keep_original,
                                                   n_rounds=n_rounds, max_labels_to_corrupt=max_labels_to_corrupt)
        self.encoding_method = label_encoder_method
        if label_encoder_method == "binary":
            self.__model_loader = BinaryYAFAL
        elif label_encoder_method == "semantic":
            self.__model_loader = SemanticYAFAL
        else:
            raise InvalidMethodException("Label encoding method should be either 'binary' or 'semantic'")

        # Training parameters
        self._val_ratio = validation_corpus_ratio
        self._max_epochs = max_epochs
        self.device = device
        self._batch_size = batch_size
        self._freeze_encoder = freeze_label_encoder
        self._freeze_sequence_encoder = freeze_sequence_encoder
        self.model = None
        self._do_early_stopping = early_stopping
        self._patience = patience
        self._best_model_state_dict = None

    def _vprint(self, text: str):
        if self._verbose:
            pprint.pprint(text)

    def fit(self, x: YAFALDataset,
            descriptors: Optional[Dict[Union[str, int], List[str]]] = None):
        """
        Parameters
        ----------

        x: YAFALDataset
            Dataset to train the Regressor

        descriptors: Dict
            Dict of the form {"label": ["describing keywords"]}

            >>> {"sports": ["football", "gym", "exercise"]}

        """

        train_losses, valid_losses = [], []
        avg_train_losses, avg_valid_losses = [], []

        # Fit the label encoder
        if self.encoding_method == "semantic":
            if not descriptors:
                raise DescriptorsRequiredException("Semantic label encoding was "
                                                   "selected but no encoding method was provided")

            self._label_encoder.fit(
                x.label_list, descriptors
            )
            if not self._label_encoder.validate_descriptions():
                raise DescriptorsRequiredException("Some of the labels do not have any descriptor")
            self.model = self.__model_loader(**self._yafal_kwargs)
            self.__freeze_label_encoder()
        else:
            self._label_encoder.fit(x.label_list, None)
            self.model = self.__model_loader(label_encoding_size=len(self._label_encoder.label_indexes),
                                             **self._yafal_kwargs)

        self.__freeze_encoder()

        # Split in train and validation sets
        x_train, x_validation = self._split_train_val(x)

        # Corrupt the datasets
        x_train_corrupt = self._dataset_corruptor.transform(
            dataset=x_train
        )
        x_val_corrupt = self._dataset_corruptor.transform(
            dataset=x_validation,
        )

        # Start the training
        torch_data_handler = YAFALTorchDataset(dataset=x_train_corrupt,
                                               label_encoder=self._label_encoder.transform if
                                               self.encoding_method == "binary" else self._label_encoder.describe,
                                               label_transform=self.encoding_method)

        torch_validation_data_handler = YAFALTorchDataset(dataset=x_val_corrupt,
                                                          label_encoder=self._label_encoder.transform if
                                                          self.encoding_method == "binary"
                                                          else self._label_encoder.describe,
                                                          label_transform=self.encoding_method)

        train_data_loader = DataLoader(
            torch_data_handler,  # The training samples.
            sampler=RandomSampler(torch_data_handler),  # Select batches randomly
            batch_size=self._batch_size  # Trains with this batch size.
        )

        val_data_loader = DataLoader(
            torch_validation_data_handler,  # The training samples.
            batch_size=self._batch_size  # Trains with this batch size.
        )

        # Use it for training
        self.model.to(torch.device(self.device))

        criterion = nn.BCELoss()  # Binary Cross Entropy Loss for 0/1 regression
        optimizer = torch.optim.AdamW(filter(lambda p: p.requires_grad, self.model.parameters()), lr=0.004)

        self.model.train()
        remaining_patience = self._patience
        best_loss = 1e6
        for epoch in range(self._max_epochs):
            self._vprint('======== Epoch {:} / {:} ========'.format(epoch + 1, self._max_epochs))

            for index, (texts, label_encodings, targets) in enumerate(train_data_loader):
                optimizer.zero_grad()
                if self.encoding_method == "semantic":
                    label_encodings = list(label_encodings)
                outputs = self.model(list(texts), label_encodings)
                loss = criterion(outputs, targets)
                train_losses.append(loss.item())
                loss.backward()
                optimizer.step()

            # Perform the validation
            self.model.eval()
            with torch.no_grad():
                for index, (texts, label_encodings, targets) in enumerate(val_data_loader):
                    if self.encoding_method == "semantic":
                        label_encodings = list(label_encodings)
                    outputs = self.model(list(texts), label_encodings)
                    val_loss = criterion(outputs, targets)
                    valid_losses.append(val_loss.item())

            avg_train_losses.append(
                np.average(train_losses)
            )
            avg_valid_losses.append(
                np.average(valid_losses)
            )
            self._vprint("Training set loss: {} / Val loss: {}".format(avg_train_losses[-1], avg_valid_losses[-1]))
            # Early Stopping
            if self._do_early_stopping:
                if avg_valid_losses[-1] > best_loss:
                    remaining_patience -= 1
                else:
                    best_loss = avg_valid_losses[-1]
                    remaining_patience = self._patience
                    self._best_model_state_dict = self.model.state_dict()
                if remaining_patience <= 0:
                    self._vprint("There were no improvements after {} rounds. Stopping.".format(self._patience))
                    self.model.load_state_dict(self._best_model_state_dict)
                    break

    def _split_train_val(self, x: YAFALDataset) -> Tuple[YAFALDataset, YAFALDataset]:
        x_train = YAFALDataset()
        x_validation = YAFALDataset()
        train_indexes, validation_indexes = x.split_indexes(ratio=self._val_ratio)
        for sample_index in train_indexes:
            sample, labels, is_corrupted = x[sample_index]
            x_train.add_sample(text=sample, sample_labels=labels, is_corrupted=is_corrupted)
        for sample_index in validation_indexes:
            sample, labels, is_corrupted = x[sample_index]
            x_validation.add_sample(text=sample, sample_labels=labels, is_corrupted=is_corrupted)
        return x_train, x_validation

    def __freeze_label_encoder(self):
        if self._freeze_encoder:
            for param in self.model.label_bert.parameters():
                param.requires_grad = False

    def __freeze_encoder(self):
        if self._freeze_sequence_encoder:
            for param in self.model.bert.parameters():
                param.requires_grad = False

    def predict(self, x: List[str], labels: List[Union[int, str, List[Union[str, int]]]]) -> np.ndarray:
        """
        Predicts score of the sample being corrupted given the list of strings and the labels of those strings

        Parameters
        ----------
        x: List[str]
            List of texts to predict

        labels:  List[Union[int, str, List[Union[str, int]]]]
            Labels associated with the input samples of x

        Returns
        -------
        np.ndarray
            Array with the predictions of shape (len(x), 1)
        """
        if self.encoding_method == "binary":
            codified_labels = self._label_encoder.transform(labels)
            label_encoding = torch.FloatTensor(codified_labels)
        else:
            label_encoding = self._label_encoder.describe(labels)
        with torch.no_grad():
            return self.model(x, label_encoding)

    def save(self, file_path: str):
        """
        Saves the model with the given filename

        Parameters
        ----------
        file_path: str
            Path to save the model
        """
        self.model.load_state_dict(self._best_model_state_dict)
        # Remove intermediate model to lighten the saved item
        self._best_model_state_dict = None
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)



import setuptools
from setuptools import find_packages

from yafal import __version__

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yafal",
    version=__version__,
    author="Manex Serras",
    author_email="mserras001@gmail.com",
    description="Identification of Fake Labels using Large Model embeddings",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://yafal.llinguai.eus",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    keywords=[
        "Natural Language Processing",
        "Artificial Intelligence",
        "Fake Label detection",
        "Deep Learning"
    ],
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=["pydantic", "numpy", "torch", "transformers", "scikit-learn"]
)



import os

from yafal import YAFALDataset, YAFALRegression
import pandas as pd

current_dir = os.path.dirname(os.path.abspath(__file__))
train_csv_path = os.path.join(current_dir, 'data', 'train.csv')
test_csv_path = os.path.join(current_dir, 'data', 'test.csv')

train_data = pd.read_csv(train_csv_path)
test_data = pd.read_csv(test_csv_path)

topic_mapping = {1: 'world',
                 2: 'sports',
                 3: 'business',
                 4: 'science'}

train_yafal_dataset = YAFALDataset()
for i, sample in train_data.iterrows():
    index_label = sample['Class Index']
    string_label = topic_mapping[index_label]
    news_article = f"{sample['Title']}. {sample['Description']}"
    train_yafal_dataset.add_sample(
        text=news_article,
        sample_labels=[index_label]
    )

# Train our model
print("Let's train our YAFAL Regressor")
yafal_regression = YAFALRegression(label_encoder_method="binary", verbose=True, max_epochs=10)
yafal_regression.fit(train_yafal_dataset)

# Predict our model
for sample in test_data.iterrows():
    result = yafal_regression.predict(["Cristiano ronaldo scores 4 goals against Manchester City"], labels=[4])
    print(result)